/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

// $Id: inst-sparc.C,v 1.97 2001/02/20 21:46:37 gurari Exp $

#include "dyninstAPI/src/inst-sparc.h"
#include "dyninstAPI/src/instPoint.h"

#include "dyninstAPI/src/FunctionExpansionRecord.h"

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

instruction newInstr[NEW_INSTR_ARRAY_LEN];
static dictionary_hash<string, unsigned> funcFrequencyTable(string::hash);

trampTemplate baseTemplate;
NonRecursiveTrampTemplate nonRecursiveBaseTemplate;

//declaration of conservative base trampoline template

trampTemplate conservativeBaseTemplate;
NonRecursiveTrampTemplate nonRecursiveConservativeBaseTemplate;



registerSpace *regSpace;

#if defined(SHM_SAMPLING) && defined(MT_THREAD)
// registers 8 to 15: out registers 
// registers 16 to 22: local registers
Register deadList[] = {16, 17, 18, 19, 20, 21, 22 };
#else
Register deadList[] = {16, 17, 18, 19, 20, 21, 22, 23 };
#endif

int deadListSize = sizeof(deadList);

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

// Constructor for the class instPoint. This one defines the 
// instPoints for the relocated function. Since the function reloated
// to the heap don't need to worry about that jump could be out of
// reach, the instructions to be moved are the one instruction at that
// point plus others if necessary(i.e. instruction in the delayed
// slot and maybe the aggregate instuction ).    
//
// Unfortunately, this constructor does NOT set several internal fields
//  e.g. size, isDelayedInsn, aggregateInsn, delaySlotInsn
instPoint::instPoint(pd_Function *f, const instruction &instr, 
                     const image *owner, Address &adr, bool delayOK, 
                     instPointType pointType, Address &oldAddr)
: addr(adr), originalInstruction(instr), inDelaySlot(false), isDelayed(false),
  callIndirect(false), callAggregate(false), callee(NULL), func(f), 
  ipType(pointType), image_ptr(owner), firstIsConditional(false), 
  relocated_(false), isLongJump(false)
{
  assert(f->isTrapFunc() == true);  

  isBranchOut = false;
  delaySlotInsn.raw = owner->get_instruction(oldAddr+4);
  aggregateInsn.raw = owner->get_instruction(oldAddr+8);

  // If the instruction is a DCTI, the instruction in the delayed 
  // slot is to be moved.
  if (IS_DELAYED_INST(instr))
    isDelayed = true;

  // If this function call another function which return an aggregate
  // value, move the aggregate instruction, too. 
  if (ipType == callSite) {
      if (!IS_VALID_INSN(aggregateInsn) && aggregateInsn.raw != 0) {
          callAggregate = true;
          adr += 8;
          oldAddr += 8;
      }
  }

  if (owner->isValidAddress(oldAddr-4)) {
    instruction iplus1;
    iplus1.raw = owner->get_instruction(oldAddr-4);
    if (IS_DELAYED_INST(iplus1) && !delayOK) {
      // ostrstream os(errorLine, 1024, ios::out);
      // os << "** inst point " << func->file->fullName << "/"
      //  << func->prettyName() << " at addr " << addr <<
      //        " in a delay slot\n";
      // logLine(errorLine);
      inDelaySlot = true;
    }
  }

  // set size at least....
  size = 0;
  if (ipType == functionEntry) {
      if (hasNoStackFrame()) {
          size = 4 * sizeof(instruction);
      } else {
          size = 5 * sizeof(instruction);
      }
  } else if (ipType == callSite) {
      size = 3 * sizeof(instruction);
  } else if (ipType == functionExit) {
      if (hasNoStackFrame()) {
          size = 5 * sizeof(instruction);
      } else {
          size = 2 * sizeof(instruction);
      }
  } else if (ipType == otherPoint){
          size = 2 * sizeof(instruction);
  } else {
      assert(false);  
  }
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

// Add the astNode opt to generate one instruction to get the 
// return value for the compiler optimazed case
void
AstNode::optRetVal(AstNode *opt) {

    if (oType == ReturnVal) {
        cout << "Optimazed Return." << endl;
        if (loperand == 0) {
            loperand = opt;
            return;
        } else if (opt == 0) {
            delete loperand;
            loperand = NULL;
            return; 
        }
    }
    if (loperand) loperand->optRetVal(opt);
    if (roperand) roperand->optRetVal(opt);
    for (unsigned i = 0; i < operands.size(); i++) 
        operands[i] -> optRetVal(opt);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

bool 
processOptimaRet(instPoint *location, AstNode *&ast) {

    // For optimazed return code
    if (location -> ipType == functionExit) {
        if ((isInsnType(location -> originalInstruction, RETmask, RETmatch)) ||
            (isInsnType(location -> originalInstruction, RETLmask, RETLmatch)))
        {
            if (isInsnType(location -> delaySlotInsn, 
                           RESTOREmask, RESTOREmatch)&&
                ((location->delaySlotInsn.raw | 0xc1e82000) != 0xc1e82000)) 
            {
                /* cout << "Optimazed Retrun Value:  Addr " << hex << 
                    location->addr << " in "
                        << location -> func -> prettyName() << endl; */
                AstNode *opt = new AstNode(AstNode::Constant,
                                           (void *)location->delaySlotInsn.raw);
                ast -> optRetVal(opt);
                return true;
            }
        }
    }
    return false;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

Register
emitOptReturn(instruction i, Register src, char *insn, Address &base, bool noCost) {
    
    unsigned instr = i.raw;

    cout << "Handling a special case for optimized return value." << endl;

    assert(((instr&0x3e000000)>>25) <= 16);

    if ((instr&0x02000)>>13)
        emitImm(plusOp, (instr&0x07c000)>>14, instr&0x01fff,
                ((instr&0x3e000000)>>25)+16, insn, base, noCost);
    else
        (void) emitV(plusOp, (instr&0x07c000)>>14, instr&0x01fff,
             ((instr&0x3e000000)>>25)+16, insn, base, noCost);
    
    return emitR(getSysRetValOp, 0, 0, src, insn, base, noCost);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

//
// initDefaultPointFrequencyTable - define the expected call frequency of
//    procedures.  Currently we just define several one shots with a
//    frequency of one, and provide a hook to read a file with more accurate
//    information.
//
void initDefaultPointFrequencyTable()
{
    FILE *fp;
    float value;
    char name[512];

    funcFrequencyTable["main"] = 1;
    funcFrequencyTable["DYNINSTsampleValues"] = 1;
    funcFrequencyTable[EXIT_NAME] = 1;

    // try to read file.
    fp = fopen("freq.input", "r");
    if (!fp) {
        return;
    } else {
        printf("found freq.input file\n");
    }
    while (!feof(fp)) {
        fscanf(fp, "%s %f\n", name, &value);
        funcFrequencyTable[name] = (int) value;
        printf("adding %s %f\n", name, value);
    }
    fclose(fp);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

/*
 * Get an estimate of the frequency for the passed instPoint.  
 *    This is not (always) the same as the function that contains the point.
 * 
 *  The function is selected as follows:
 *
 *  If the point is an entry or an exit return the function name.
 *  If the point is a call and the callee can be determined, return the called
 *     function.
 *  else return the funcation containing the point.
 *
 *  WARNING: This code contins arbitray values for func frequency (both user 
 *     and system).  This should be refined over time.
 *
 * Using 1000 calls sec to be one SD from the mean for most FPSPEC apps.
 *      -- jkh 6/24/94
 *
 */
float getPointFrequency(instPoint *point)
{

    pd_Function *func;

    if (point->callee)
        func = point->callee;
    else
        func = point->func;

    if (!funcFrequencyTable.defines(func->prettyName())) {
      // Changing this value from 250 to 100 because predictedCost was
      // too high - naim 07/18/96
      return(100); 
      
    } else {
      return (funcFrequencyTable[func->prettyName()]);
    }
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

//
// return cost in cycles of executing at this point.  This is the cost
//   of the base tramp if it is the first at this point or 0 otherwise.
//
int getPointCost(process *proc, const instPoint *point)
{
    if (proc->baseMap.defines(point)) {
        return(0);
    } else {
        // 70 cycles for base tramp (worst case)
        return(70);
    }
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void initATramp (trampTemplate *thisTemp, Address tramp,
		 bool isConservative=false)
{
    instruction *temp;

    if(!isConservative){

    	thisTemp->savePreInsOffset = 
			((Address)baseTramp_savePreInsn - tramp);
    	thisTemp->restorePreInsOffset = 
			((Address)baseTramp_restorePreInsn - tramp);
    	thisTemp->savePostInsOffset = 
			((Address)baseTramp_savePostInsn - tramp);
    	thisTemp->restorePostInsOffset = 
			((Address)baseTramp_restorePostInsn - tramp);
    }else{
	thisTemp->savePreInsOffset =
		((Address)conservativeBaseTramp_savePreInsn - tramp);
	thisTemp->restorePreInsOffset =
		((Address)conservativeBaseTramp_restorePreInsn - tramp);
	thisTemp->savePostInsOffset =
		((Address)conservativeBaseTramp_savePostInsn - tramp);
	thisTemp->restorePostInsOffset =
		((Address)conservativeBaseTramp_restorePostInsn - tramp);
    }

    // TODO - are these offsets always positive?
    thisTemp->trampTemp = (void *) tramp;
    for (temp = (instruction*)tramp; temp->raw != END_TRAMP; temp++) {
        const Address offset = (Address)temp - tramp;
        switch (temp->raw) {
            case LOCAL_PRE_BRANCH:
                thisTemp->localPreOffset = offset;
                thisTemp->localPreReturnOffset = thisTemp->localPreOffset 
                                                + sizeof(temp->raw);
                break;
            case GLOBAL_PRE_BRANCH:
                thisTemp->globalPreOffset = offset;
                break;
            case LOCAL_POST_BRANCH:
                thisTemp->localPostOffset = offset;
                thisTemp->localPostReturnOffset = thisTemp->localPostOffset
                                                + sizeof(temp->raw);
                break;
            case GLOBAL_POST_BRANCH:
                thisTemp->globalPostOffset = offset;
                break;
            case SKIP_PRE_INSN:
                thisTemp->skipPreInsOffset = offset;
                break;
            case UPDATE_COST_INSN:
                thisTemp->updateCostOffset = offset;
                break;
            case SKIP_POST_INSN:
                thisTemp->skipPostInsOffset = offset;
                break;
            case RETURN_INSN:
                thisTemp->returnInsOffset = offset;
                break;
            case EMULATE_INSN:
                thisTemp->emulateInsOffset = offset;
                break;
	    case CONSERVATIVE_TRAMP_READ_CONDITION:
		if(isConservative)
			temp->raw = 0x83408000; /*read condition codes to g1*/
		break;
	    case CONSERVATIVE_TRAMP_WRITE_CONDITION:
		if(isConservative)
			temp->raw = 0x85806000; /*write condition codes fromg1*/
		break;
        }       
    }

    // Cost with the skip branches.
    thisTemp->cost = 14;  
    thisTemp->prevBaseCost = 20 +
      RECURSIVE_GUARD_ON_CODE_SIZE + RECURSIVE_GUARD_OFF_CODE_SIZE;
    thisTemp->postBaseCost = 22 +
      RECURSIVE_GUARD_ON_CODE_SIZE + RECURSIVE_GUARD_OFF_CODE_SIZE;
    thisTemp->prevInstru = thisTemp->postInstru = false;
    thisTemp->size = (int) temp - (int) tramp;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void initATramp(NonRecursiveTrampTemplate *thisTemp, Address tramp,
		bool isConservative=false)
{
  initATramp((trampTemplate *)thisTemp, tramp,isConservative);

  instruction *temp;

  for (temp = (instruction*)tramp; temp->raw != END_TRAMP; temp++) {
    const Address offset = (Address)temp - tramp;
    switch (temp->raw)
      {

      case RECURSIVE_GUARD_ON_PRE_INSN:
	thisTemp->guardOnPre_beginOffset = offset;
	thisTemp->guardOnPre_endOffset = thisTemp->guardOnPre_beginOffset
                + RECURSIVE_GUARD_ON_CODE_SIZE * INSN_SIZE;
	break;

      case RECURSIVE_GUARD_OFF_PRE_INSN:
	thisTemp->guardOffPre_beginOffset = offset;
	thisTemp->guardOffPre_endOffset = thisTemp->guardOffPre_beginOffset
                + RECURSIVE_GUARD_OFF_CODE_SIZE * INSN_SIZE;
	break;

      case RECURSIVE_GUARD_ON_POST_INSN:
	thisTemp->guardOnPost_beginOffset = offset;
	thisTemp->guardOnPost_endOffset = thisTemp->guardOnPost_beginOffset
                + RECURSIVE_GUARD_ON_CODE_SIZE * INSN_SIZE;
	break;

      case RECURSIVE_GUARD_OFF_POST_INSN:
	thisTemp->guardOffPost_beginOffset = offset;
	thisTemp->guardOffPost_endOffset = thisTemp->guardOffPost_beginOffset
                + RECURSIVE_GUARD_OFF_CODE_SIZE * INSN_SIZE;
	break;

      }
  }
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void initTramps()
{
    static bool inited=false;

    if (inited) return;
    inited = true;

    initATramp(&baseTemplate, (Address) baseTramp);
    initATramp(&nonRecursiveBaseTemplate, (Address)baseTramp);


    initATramp(&conservativeBaseTemplate,
	       (Address)conservativeBaseTramp,true);
    initATramp(&nonRecursiveConservativeBaseTemplate,
	       (Address)conservativeBaseTramp,true);

    regSpace = new registerSpace(sizeof(deadList)/sizeof(Register), deadList,
                                         0, NULL);
    assert(regSpace);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

#if defined(MT_THREAD)
//
// For multithreaded applications and shared memory sampling, this routine 
// will compute the address where the corresponding counter/timer vector for
// level 0 is (by default). In the mini-tramp, if the counter/timer is at a
// different level, we will add the corresponding offset - naim 4/18/97
//
// NUM_INSN_MT_PREAMBLE
void generateMTpreamble(char *insn, Address &base, process *proc)
{
  AstNode *t1 ;
  Address tableAddr;
  int value; 
  bool err;
  unsigned offset = 19*sizeof(instruction);
  Register src = Null_Register;

  vector<AstNode *> dummy;

  t1 = new AstNode("DYNINSTthreadPos", dummy);

  // t2=DYNINSTthreadPos()*sizeof(unsigned)
  value = sizeof(unsigned);
  AstNode* t4 = new AstNode(AstNode::Constant,(void *)value);
  AstNode* t2 = new AstNode(timesOp, t1, t4);
  removeAst(t1) ;
  removeAst(t4) ;

  // t3=DYNINSTthreadTable+t2
  //
  tableAddr = proc->findInternalAddress("DYNINSTthreadTable",true,err);
  assert(!err);
  AstNode* t5 = new AstNode(AstNode::Constant, (void *)tableAddr);
  AstNode* t3 = new AstNode(plusOp, t2, t5);
  removeAst(t2);
  removeAst(t5);

  AstNode* t7 = new AstNode(AstNode::Constant,(void *)-2);
  AstNode* t8 = new AstNode(eqOp, t1, t7);
  removeAst(t7);
  AstNode* t9 = new AstNode(AstNode::Constant, (void *)(offset));
  //goto offset
  AstNode* t10 = new AstNode(branchOp, t9);
  removeAst(t9);
  //t11 = "if(t1 ==-2) goto offset"
  AstNode* t11 = new AstNode(ifOp, t8, t10);
  removeAst(t8);
  removeAst(t10) ;

  //t6 = " if (t1 == -2) goto Offset; t3=DYNINSTthreadTable+t2"
  AstNode *t6= new AstNode(t11, t3);
  removeAst(t11);
  removeAst(t3);
  src = t6->generateCode(proc, regSpace, insn, base, false, true);
  removeAst(t6);
  (void) emitV(orOp, src, 0, REG_MT, insn, base, false);
  regSpace->freeRegister(src);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void generateRPCpreamble(char *insn, Address &base, process *proc, unsigned offset, int tid, unsigned pos)
{
  AstNode *t1 ;
  Address tableAddr;
  int value; 
  bool err;
  Register src = Null_Register;

  if (tid != -1)  {
    vector<AstNode *> param;
    param += new AstNode(AstNode::Constant,(void *)tid);
    param += new AstNode(AstNode::Constant,(void *)pos);
    t1 = new AstNode("DYNINSTthreadPosTID", param);
    for (unsigned i=0; i<param.size(); i++) 
      removeAst(param[i]) ;

    value = sizeof(unsigned);
    AstNode* t4 = new AstNode(AstNode::Constant,(void *)value);
    AstNode* t2 = new AstNode(timesOp, t1, t4);
    removeAst(t1) ;
    removeAst(t4) ;

    // t3=DYNINSTthreadTable[thr_self()]
    //
    tableAddr = proc->findInternalAddress("DYNINSTthreadTable",true,err);
    assert(!err);
    AstNode* t5 = new AstNode(AstNode::Constant, (void *)tableAddr);
    AstNode* t3 = new AstNode(plusOp, t2, t5);
    removeAst(t2);
    removeAst(t5);

    AstNode* t7 = new AstNode(AstNode::Constant,(void *)-2);
    AstNode* t8 = new AstNode(eqOp, t1, t7);
    removeAst(t7);
    AstNode* t9 = new AstNode(AstNode::Constant, (void *)(offset));
    AstNode* t10 = new AstNode(branchOp, t9);
    removeAst(t9);
    AstNode* t11 = new AstNode(ifOp, t8, t10);
    removeAst(t8);
    removeAst(t10);

    AstNode *t6= new AstNode(t11, t3);
    removeAst(t11);
    removeAst(t3);
    src = t6->generateCode(proc, regSpace, insn, base, false, true);
    removeAst(t6);
    unsigned first_insn=base;
    (void) emitV(orOp, src, 0, REG_MT, insn, base, false);
    regSpace->freeRegister(src);
  }
}
#endif

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void generateNoOp(process *proc, Address addr)
{
    instruction insn;

    /* fill with no-op */
    insn.raw = 0;
    insn.branch.op = 0;
    insn.branch.op2 = NOOPop2;

    proc->writeTextWord((caddr_t)addr, insn.raw);
}


/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

/*
 * change the insn at addr to be a branch to newAddr.
 *   Used to add multiple tramps to a point.
 */
void generateBranch(process *proc, Address fromAddr, Address newAddr)
{
    int disp;
    instruction insn;

    disp = newAddr-fromAddr;
    generateBranchInsn(&insn, disp);

    proc->writeTextWord((caddr_t)fromAddr, insn.raw);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void generateCall(process *proc, Address fromAddr, Address newAddr)
{
    instruction insn; 
    generateCallInsn(&insn, fromAddr, newAddr);

    proc->writeTextWord((caddr_t)fromAddr, insn.raw);

}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void genImm(process *proc, Address fromAddr, int op, 
                        Register rs1, int immd, Register rd)
{
    instruction insn;
    genImmInsn(&insn, op, rs1, immd, rd);

    proc->writeTextWord((caddr_t)fromAddr, insn.raw);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

/*
 *  change the target of the branch at fromAddr, to be newAddr.
 */
void changeBranch(process *proc, Address fromAddr, Address newAddr, 
                  instruction originalBranch) {
    int disp = newAddr-fromAddr;
    instruction insn;
    insn.raw = originalBranch.raw;
    insn.branch.disp22 = disp >> 2;
    proc->writeTextWord((caddr_t)fromAddr, insn.raw);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

int callsTrackedFuncP(instPoint *point)
{
    if (point->callIndirect) {
        return(true);
    } else {
        if (point->callee) {
            return(true);
        } else {
            return(false);
        }
    }
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

/*
 * return the function asociated with a point.
 *
 *     If the point is a funcation call, and we know the function being called,
 *          then we use that.  Otherwise it is the function that contains the
 *          point.
 *  
 *   This is done to return a better idea of which function we are using.
 */
pd_Function *getFunction(instPoint *point)
{
    return(point->callee ? point->callee : point->func);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

bool process::emitInferiorRPCheader(void *insnPtr, Address &baseBytes) {
   instruction *insn = (instruction *)insnPtr;
   Address baseInstruc = baseBytes / sizeof(instruction);

   genImmInsn(&insn[baseInstruc++], SAVEop3, 14, -112, 14);

   baseBytes = baseInstruc * sizeof(instruction); // convert back
   return true;
}


/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

bool process::emitInferiorRPCtrailer(void *insnPtr, Address &baseBytes,
                                     unsigned &breakOffset,
                                     bool stopForResult,
                                     unsigned &stopForResultOffset,
#if defined(MT_THREAD)
				     unsigned &justAfter_stopForResultOffset,
				     bool isSafeRPC) {
#else
                                     unsigned &justAfter_stopForResultOffset) {
#endif
   // Sequence: restore, trap, illegal

   instruction *insn = (instruction *)insnPtr;
   Address baseInstruc = baseBytes / sizeof(instruction);

   if (stopForResult) {
      // trap insn:
      genBreakpointTrap(&insn[baseInstruc]);
      stopForResultOffset = baseInstruc * sizeof(instruction);
      baseInstruc++;
      justAfter_stopForResultOffset = baseInstruc * sizeof(instruction);
   }

#if defined(MT_THREAD)
   if (isSafeRPC) //ret instruction
     genImmInsn(&insn[baseInstruc++], JMPLop3, REG_I(7), 0x08, REG_G(0)) ;
#endif
   genSimpleInsn(&insn[baseInstruc++], RESTOREop3, 0, 0, 0);

   // Now that the inferior has executed the 'restore' instruction, the %in and
   // %local registers have been restored.  We mustn't modify them after this point!!
   // (reminder: the %in and %local registers aren't saved and set with ptrace
   //  GETREGS/SETREGS call)

#if defined(MT_THREAD)
   if (!isSafeRPC) {
#endif
     // Trap instruction:
     genBreakpointTrap(&insn[baseInstruc]); // ta 1
     breakOffset = baseInstruc * sizeof(instruction);
     baseInstruc++;

     // And just to make sure that we don't continue from the trap:
     genUnimplementedInsn(&insn[baseInstruc++]); // UNIMP 0
#if defined(MT_THREAD)
   }
#endif
   baseBytes = baseInstruc * sizeof(instruction); // convert back

   return true; // success
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void emitImm(opCode op, Register src1, RegValue src2imm, Register dest, 
             char *i, Address &base, bool noCost)
{
        instruction *insn = (instruction *) ((void*)&i[base]);
        RegValue op3 = -1;
        int result = -1;
        switch (op) {
            // integer ops
            case plusOp:
                op3 = ADDop3;
                genImmInsn(insn, op3, src1, src2imm, dest);
                break;

            case minusOp:
                op3 = SUBop3;
                genImmInsn(insn, op3, src1, src2imm, dest);
                break;

            case timesOp:
                op3 = SMULop3;
                if (isPowerOf2(src2imm,result) && (result<32))
                  generateLShift(insn, src1, (Register)result, dest);           
                else 
                  genImmInsn(insn, op3, src1, src2imm, dest);
                break;

            case divOp:
                op3 = SDIVop3;
                if (isPowerOf2(src2imm,result) && (result<32))
                  generateRShift(insn, src1, (Register)result, dest);           
                else { // needs to set the Y register to zero first
                  // Set the Y register to zero: Zhichen
                  genImmInsn(insn, WRYop3, REG_G(0), 0, 0);
                  base += sizeof(instruction);
                  insn = (instruction *) ((void*)&i[base]);
                  genImmInsn(insn, op3, src1, src2imm, dest);
                }

                break;

            // Bool ops
            case orOp:
                op3 = ORop3;
                genImmInsn(insn, op3, src1, src2imm, dest);
                break;

            case andOp:
                op3 = ANDop3;
                genImmInsn(insn, op3, src1, src2imm, dest);
                break;

            // rel ops
            // For a particular condition (e.g. <=) we need to use the
            // the opposite in order to get the right value (e.g. for >=
            // we need BLTcond) - naim
            case eqOp:
                genImmRelOp(insn, BNEcond, src1, src2imm, dest, base);
                return;
                break;

            case neOp:
                genImmRelOp(insn, BEcond, src1, src2imm, dest, base);
                return;
                break;

            case lessOp:
                genImmRelOp(insn, BGEcond, src1, src2imm, dest, base);
                return;
                break;

            case leOp:
                genImmRelOp(insn, BGTcond, src1, src2imm, dest, base);
                return;
                break;

            case greaterOp:
                genImmRelOp(insn, BLEcond, src1, src2imm, dest, base);
                return;
                break;

            case geOp:
                genImmRelOp(insn, BLTcond, src1, src2imm, dest, base);
                return;
                break;

            default:
                Register dest2 = regSpace->allocateRegister(i, base, noCost);
                (void) emitV(loadConstOp, src2imm, dest2, dest2, i, base, noCost);
                (void) emitV(op, src1, dest2, dest, i, base, noCost);
                regSpace->freeRegister(dest2);
                return;
                break;
        }
        base += sizeof(instruction);
        return;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

//
// All values based on Cypress0 && Cypress1 implementations as documented in
//   SPARC v.8 manual p. 291
//
int getInsnCost(opCode op)
{
    /* XXX Need to add branchOp */
    if (op == loadConstOp) {
        return(1);
    } else if (op ==  loadOp) {
        // sethi + load single
        return(1+1);
    } else if (op ==  loadIndirOp) {
        return(1);
    } else if (op ==  storeOp) {
        // sethi + store single
        // return(1+3); 
        // for SS-5 ?
        return(1+2); 
    } else if (op ==  storeIndirOp) {
        return(2); 
    } else if (op ==  ifOp) {
        // subcc
        // be
        // nop
        return(1+1+1);
    } else if (op ==  callOp) {
        int count = 0;

        // mov src1, %o0
        count += 1;

        // mov src2, %o1
        count += 1;

        // clr i2
        count += 1;

        // clr i3
        count += 1;

        // sethi
        count += 1;

        // jmpl
        count += 1;

        // noop
        count += 1;

        return(count);
    } else if (op ==  updateCostOp) {
        // sethi %hi(obsCost), %l0
        // ld [%lo + %lo(obsCost)], %l1
        // add %l1, <cost>, %l1
        // st %l1, [%lo + %lo(obsCost)]
        return(1+2+1+3);
    } else if (op ==  trampPreamble) {
        return(0);
    } else if (op ==  trampTrailer) {
        // retl
        return(2);
    } else if (op == noOp) {
        // noop
        return(1);
    } else if (op == getParamOp) {
        return(0);
    } else {
        switch (op) {
            // rel ops
            case eqOp:
            case neOp:
            case lessOp:
            case leOp:
            case greaterOp:
            case geOp:
                // bne -- assume taken
                return(2);
                break;
            default:
                return(1);
                break;
        }
    }
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

bool isReturnInsn(instruction instr, Address addr, string name) {
    if (isInsnType(instr, RETmask, RETmatch) ||
        isInsnType(instr, RETLmask, RETLmatch)) {
        //  Why 8 or 12?
        //  According to the sparc arch manual (289), ret or retl are
        //   synthetic instructions for jmpl %i7+8, %g0 or jmpl %o7+8, %go.
        //  Apparently, the +8 is not really a hard limit though, as 
        //   sometimes some extra space is allocated after the jump 
        //   instruction for various reasons.
        //  1 possible reason is to include information on the size of
        //   returned structure (4 bytes).
        //  So, 8 or 12 here is a heuristic, but doesn't seem to 
        //   absolutely have to be true.
        //  -matt
        if ((instr.resti.simm13 != 8) && (instr.resti.simm13 != 12) 
                    && (instr.resti.simm13 != 16)) {
          sprintf(errorLine,"WARNING: unsupported return at address 0x%lx"
                        " in function %s - appears to be return to PC + %i", 
                  addr, name.string_of(), (int)instr.resti.simm13);
          showErrorCallback(55, errorLine);
        } else { 
          return true;
        }
    }
    return false;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

bool isReturnInsn(const image *owner, Address adr, bool &lastOne, string name)
{
    instruction instr;

    instr.raw = owner->get_instruction(adr);
    lastOne = false;

    return isReturnInsn(instr, adr, name);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

bool isBranchInsn(instruction instr) {
    if (instr.branch.op == 0 
                && (instr.branch.op2 == 2 || instr.branch.op2 == 6)) 
          return true;
    return false;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

// modifyInstPoint: if the function associated with the process was 
// recently relocated, then the instPoint may have the old pre-relocated
// address (this can occur because we are getting instPoints in mdl routines 
// and passing these to routines that do the instrumentation, it would
// be better to let the routines that do the instrumenting find the points)
void pd_Function::modifyInstPoint(const instPoint *&location,process *proc)
{

    if(relocatable_ && !(location->relocated_)){
        for(u_int i=0; i < relocatedByProcess.size(); i++){
            if((relocatedByProcess[i])->getProcess() == proc){
                if(location->ipType == functionEntry){
                    const instPoint *new_entry = 
                                (relocatedByProcess[i])->funcEntry();
                    location = new_entry;
                } 
                else if(location->ipType == functionExit){
                    const vector<instPoint *> new_returns = 
                        (relocatedByProcess[i])->funcReturns(); 
                    if(funcReturns.size() != new_returns.size()){
                        printf("funcReturns = %d new_returns = %d\n",
                                funcReturns.size(),new_returns.size());
                        fflush(stdout);
                    }
                    assert(funcReturns.size() == new_returns.size());
                    for(u_int j=0; j < new_returns.size(); j++){
                        if(funcReturns[j] == location){
                            location = (new_returns[j]);
                            break;
                        }
                    }
                }
                else {
                    const vector<instPoint *> new_calls = 
                                (relocatedByProcess[i])->funcCallSites(); 
                    assert(calls.size() == new_calls.size());
                    for(u_int j=0; j < new_calls.size(); j++){
                        if(calls[j] == location){
                            location = (new_calls[j]);
                            break;
                        }
                    }

                }
                break;
    } } }
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

// The exact semantics of the heap are processor specific.
//
// find all DYNINST symbols that are data symbols
//
bool process::heapIsOk(const vector<sym_data> &find_us) {
  Symbol sym;
  Address baseAddr;

  // find the main function
  // first look for main or _main
  if (!((mainFunction = findOneFunction("main")) 
        || (mainFunction = findOneFunction("_main")))) {
     string msg = "Cannot find main. Exiting.";
     statusLine(msg.string_of());
#if defined(BPATCH_LIBRARY)
     BPatch_reportError(BPatchWarning, 50, msg.string_of());
#else
     showErrorCallback(50, msg);
#endif
     return false;
  }

  for (unsigned i=0; i<find_us.size(); i++) {
    const string &str = find_us[i].name;
    if (!getSymbolInfo(str, sym, baseAddr)) {
      string str1 = string("_") + str.string_of();
      if (!getSymbolInfo(str1, sym, baseAddr) && find_us[i].must_find) {
        string msg;
        msg = string("Cannot find ") + str + string(". Exiting");
        statusLine(msg.string_of());
        showErrorCallback(50, msg);
        return false;
      }
    }
  }

//  string ghb = GLOBAL_HEAP_BASE;
//  if (!getSymbolInfo(ghb, sym, baseAddr)) {
//    ghb = U_GLOBAL_HEAP_BASE;
//    if (!linkedFile.get_symbol(ghb, sym)) {
//      string msg;
//      msg = string("Cannot find ") + ghb + string(". Exiting");
//      statusLine(msg.string_of());
//      showErrorCallback(50, msg);
//      return false;
//    }
//  }
//  Address instHeapEnd = sym.addr()+baseAddr;
//  addInternalSymbol(ghb, instHeapEnd);


#ifdef ndef
  /* Not needed with the new heap type system */

  string ihb = INFERIOR_HEAP_BASE;
  if (!getSymbolInfo(ihb, sym, baseAddr)) {
    ihb = UINFERIOR_HEAP_BASE;
    if (!getSymbolInfo(ihb, sym, baseAddr)) {
      string msg;
      msg = string("Cannot find ") + ihb + string(". Cannot use this application");
      statusLine(msg.string_of());
      showErrorCallback(50, msg);
      return false;
    }
  }

  Address curr = sym.addr()+baseAddr;
  // Check that we can patch up user code to jump to our base trampolines
  // (Perhaps this code is no longer needed for sparc platforms, since we use full
  // 32-bit jumps)
  const Address instHeapStart = curr;
  const Address instHeapEnd = instHeapStart + SYN_INST_BUF_SIZE - 1;

  if (instHeapEnd > getMaxBranch3Insn()) {
    logLine("*** FATAL ERROR: Program text + data too big for dyninst\n");
    sprintf(errorLine, "    heap starts at %x and ends at %x; maxbranch=%x\n",
            instHeapStart, instHeapEnd, getMaxBranch3Insn());
    logLine(errorLine);
    return false;
  }
#endif

  return true;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

// Certain registers (i0-i7 on a SPARC) may be available to be read
// as an operand, but cannot be written.
bool registerSpace::readOnlyRegister(Register /*reg_number*/) {
// -- this code removed, since it seems incorrect
//if ((reg_number < REG_L(0)) || (reg_number > REG_L(7)))
//    return true;
//else
      return false;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

bool returnInstance::checkReturnInstance(const vector<Address> &stack, u_int &index) {
    // If false (unsafe) is returned, then 'index' is set to the first unsafe call stack
    // index.

  //    cout << "checkReturnInstance: addr_=" << (void*)addr_ << "; size_=" <<
  //            (void*)size_ << endl;
  //    cout << "instruction sequence is:" << endl;
  //    for (unsigned i=0; i < instSeqSize/4; i ++)
  //       cout << (void*)instructionSeq[i].raw << endl;
  //    cout << endl;
  //    
  //    for (unsigned i=0; i < stack.size(); i++)
  //       cout << (void*)stack[i] << endl;

    for (u_int i=0; i < stack.size(); i++) {
        index = i;

        // Is the following check correct?  Shouldn't the ">" be changed to ">=",
        // and the "<=" be changed to "<" ??? --ari 6/11/97
        if (stack[i] > addr_ && stack[i] <= addr_+size_)
            return false;
    }

    return true;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void returnInstance::installReturnInstance(process *proc) {
    proc->writeTextSpace((caddr_t)addr_, instSeqSize, 
                         (caddr_t) instructionSeq); 
	installed = true;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void generateBreakPoint(instruction &insn) {
    insn.raw = BREAK_POINT_INSN;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void returnInstance::addToReturnWaitingList(Address pc, process *proc) {
    // if there already is a TRAP set at this pc for this process don't
    // generate a trap instruction again...you will get the wrong original
    // instruction if you do a readDataSpace
    bool found = false;
    instruction insn;
    for (u_int i=0; i < instWList.size(); i++) {
         if (instWList[i]->pc_ == pc && instWList[i]->which_proc == proc) {
             found = true;
             insn = instWList[i]->relocatedInstruction;
             break;
         }
    }
    if(!found) {
        instruction insnTrap;
        generateBreakPoint(insnTrap);
        proc->readDataSpace((caddr_t)pc, sizeof(insn), (char *)&insn, true);
        proc->writeTextSpace((caddr_t)pc, sizeof(insnTrap), (caddr_t)&insnTrap);
    }
    else {
    }

    instWaitingList *instW = new instWaitingList(instructionSeq,instSeqSize,
                                                 addr_,pc,insn,pc,proc);
    instWList += instW;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

bool doNotOverflow(int value)
{
  // we are assuming that we have 13 bits to store the immediate operand.
  //if ( (value <= 16383) && (value >= -16384) ) return(true);
  if ( (value <= MAX_IMM13) && (value >= MIN_IMM13) ) return(true);
  else return(false);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

void instWaitingList::cleanUp(process *proc, Address pc) {
    proc->writeTextSpace((caddr_t)pc, sizeof(relocatedInstruction),
                    (caddr_t)&relocatedInstruction);
    proc->writeTextSpace((caddr_t)addr_, instSeqSize, (caddr_t)instructionSeq);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

// process::replaceFunctionCall
//
// Replace the function call at the given instrumentation point with a call to
// a different function, or with a NOOP.  In order to replace the call with a
// NOOP, pass NULL as the parameter "func."
// Returns true if sucessful, false if not.  Fails if the site is not a call
// site, or if the site has already been instrumented using a base tramp.
bool process::replaceFunctionCall(const instPoint *point,
                                  const function_base *func) {
    // Must be a call site
    if (point->ipType != callSite)
        return false;

    // Cannot already be instrumented with a base tramp
    if (baseMap.defines(point))
        return false;

    // Replace the call
    if (func == NULL)
        generateNoOp(this, point->addr);
    else
        generateCall(this, point->addr,
                     func->getEffectiveAddress(this));

    return true;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

#ifndef BPATCH_LIBRARY
bool process::isDynamicCallSite(instPoint *callSite){
  function_base *temp;
  if(!findCallee(*(callSite),temp)){
    //True call instructions are not dynamic on sparc,
    //they are always to a pc relative offset
    if(!isTrueCallInsn(callSite->originalInstruction))
      return true;
  }
  return false;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

bool process::MonitorCallSite(instPoint *callSite){
 
  if(isJmplInsn(callSite->originalInstruction)){
    vector<AstNode *> the_args(2);
    
    //this instruction is a jmpl with i == 1, meaning it
    //calling function register rs1+simm13
    if(callSite->originalInstruction.rest.i == 1){
      
      AstNode *base =  new AstNode(AstNode::PreviousStackFrameDataReg,
			  (void *) callSite->originalInstruction.rest.rs1);
      AstNode *offset = new AstNode(AstNode::Constant, 
			(void *) callSite->originalInstruction.resti.simm13);
      the_args[0] = new AstNode(plusOp, base, offset);
    } 
    
    //This instruction is a jmpl with i == 0, meaning its
    //two operands are registers
    else if(callSite->originalInstruction.rest.i == 0){
      //Calculate the byte offset from the contents of the %fp reg
      //that the registers from the previous stack frame 
      //specified by rs1 and rs2 are stored on the stack
      AstNode *callee_addr1 = 
	new AstNode(AstNode::PreviousStackFrameDataReg,
		    (void *) callSite->originalInstruction.rest.rs1);
      AstNode *callee_addr2 = 
	new AstNode(AstNode::PreviousStackFrameDataReg, 
		    (void *) callSite->originalInstruction.rest.rs2);
      the_args[0] = new AstNode(plusOp, callee_addr1, callee_addr2);
    }
    else assert(0);
    
    the_args[1] = new AstNode(AstNode::Constant,
			      (void *) callSite->iPgetAddress());
    AstNode *func = new AstNode("DYNINSTRegisterCallee", 
				the_args);
    addInstFunc(this, callSite, func, callPreInsn,
		orderFirstAtPoint,
		true,
		false);
  }
  else if(isTrueCallInsn(callSite->originalInstruction)){
    //True call destinations are always statically determinable.
    return true;
  }
  else return false;

  return true;
}
#endif

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

// Emit code to jump to function CALLEE without linking.  (I.e., when
// CALLEE returns, it returns to the current caller.)  On SPARC, we do
// this by ensuring that the register context upon entry to CALLEE is
// the register context of function we are instrumenting, popped once.
void emitFuncJump(opCode op, 
		  char *i, Address &base, 
		  const function_base *callee,
		  process *proc)
{
        assert(op == funcJumpOp);
        Address addr;
	void cleanUpAndExit(int status);

	addr = callee->getEffectiveAddress(proc);
	// TODO cast
	instruction *insn = (instruction *) ((void*)&i[base]);

        generateSetHi(insn, addr, 13); insn++;
	// don't want the return address to be used
        genImmInsn(insn, JMPLop3, 13, LOW10(addr), 0); insn++;
        genSimpleInsn(insn, RESTOREop3, 0, 0, 0); insn++;
        base += 3 * sizeof(instruction);
}

#ifdef BPATCH_LIBRARY

#include "BPatch_flowGraph.h"

#include <sys/systeminfo.h>

/*
 * function which check whether the architecture is 
 * sparcv8plus or not. For the earlier architectures 
 * it is not possible to support random instrumentation 
 */
bool isV8plusISA()
{
    char isaOptions[256];

    if (sysinfo(SI_ISALIST, isaOptions, 256) < 0)
        return false;
    if (strstr(isaOptions, "sparcv8plus"))
        return true;
    return false;
}

/*
 * createInstructionInstPoint
 *
 * Create a BPatch_point instrumentation point at the given address, which
 * is guaranteed not be one of the "standard" inst points.
 *
 * proc         The process in which to create the inst point.
 * address      The address for which to create the point.
 */
BPatch_point *createInstructionInstPoint(process *proc, void *address)
{
    unsigned i;
    Address begin_addr,end_addr,curr_addr = (Address)address;

    //the method to check whether conservative base tramp can be installed
    //or not since it contains condition code instructions which is
    //available after version8plus of sparc

    if(!isV8plusISA()){
	cerr << "BPatch_image::createInstPointAtAddr : is not supported for";
	cerr << " sparc architecture earlier than v8plus\n";
	return NULL;
    }

    //if the address is not aligned then there is a problem
    if(!isAligned(curr_addr))	
	return NULL;

    function_base *func = proc->findFunctionIn(curr_addr);

    if (func != NULL) {
	instPoint *entry = const_cast<instPoint *>(func->funcEntry(proc));
	assert(entry);

	begin_addr = entry->iPgetAddress();
	end_addr = begin_addr + entry->Size();

	if(((begin_addr - INSN_SIZE) <= curr_addr) && 
	   (curr_addr < end_addr)){ 
	    BPatch_reportError(BPatchSerious, 117,
			       "instrumentation point conflict");
	    return NULL;
	}

	const vector<instPoint*> &exits = func->funcExits(proc);
	for (i = 0; i < exits.size(); i++) {
	    assert(exits[i]);

	    begin_addr = exits[i]->iPgetAddress();
	    end_addr = begin_addr + exits[i]->Size();

	    if (((begin_addr - INSN_SIZE) <= curr_addr) &&
		(curr_addr < end_addr)){
		BPatch_reportError(BPatchSerious, 117,
				   "instrumentation point conflict");
		return NULL;
	    }
	}

	const vector<instPoint*> &calls = func->funcCalls(proc);
	for (i = 0; i < calls.size(); i++) {
	    assert(calls[i]);

	    begin_addr = calls[i]->iPgetAddress();
	    end_addr = begin_addr + calls[i]->Size();

	    if (((begin_addr - INSN_SIZE) <= curr_addr) &&
		(curr_addr < end_addr)){
		BPatch_reportError(BPatchSerious, 117,
				   "instrumentation point conflict");
		return NULL;
	    }
	}
    }

    /* Check for conflict with a previously created inst point. */
    if (proc->instPointMap.defines(curr_addr - INSN_SIZE)) {
	BPatch_reportError(BPatchSerious,117,"instrumentation point conflict");
	return NULL;
    } else if (proc->instPointMap.defines(curr_addr + INSN_SIZE)) {
	BPatch_reportError(BPatchSerious,117,"instrumentation point conflict");
	return NULL;
    }

    /* Check for instrumentation where the delay slot of the jump to the
       base tramp would be a branch target from elsewhere in the function. */

    BPatch_function *bpfunc = proc->PDFuncToBPFuncMap[func];
    /* XXX Should create here with correct module, but we don't know it. */
    if (bpfunc == NULL) bpfunc = new BPatch_function(proc, func, NULL);

    BPatch_flowGraph *cfg = bpfunc->getCFG();
    BPatch_basicBlock** belements =
                new BPatch_basicBlock*[cfg->getAllBasicBlocks()->size()];
    cfg->getAllBasicBlocks()->elements(belements);

    for(i=0; i< (unsigned)cfg->getAllBasicBlocks()->size(); i++) {
	void *bbsa, *bbea;
	if (belements[i]->getAddressRange(bbsa,bbea)) {
	    begin_addr = (Address)bbsa;
	    if ((begin_addr - INSN_SIZE) == curr_addr) {
		delete[] belements;
		BPatch_reportError(BPatchSerious, 118,
				   "point uninstrumentable");
		return NULL;
	    }
	}
    }
    delete[] belements;

    /* Check for instrumenting just before or after a branch. */

    if (curr_addr > func->getEffectiveAddress(proc)) {
	instruction prevInstr;
	proc->readTextSpace((char *)address - INSN_SIZE,
			    sizeof(instruction),
			    &prevInstr.raw);
	if (IS_DELAYED_INST(prevInstr)){
	    BPatch_reportError(BPatchSerious, 118, "point uninstrumentable");
	    return NULL;
	}
    }

    if ((Address)address + INSN_SIZE <
	func->getEffectiveAddress(proc) + func->size()) {
	instruction nextInstr;
	proc->readTextSpace((char *)address + INSN_SIZE,
			    sizeof(instruction),
			    &nextInstr.raw);
	if (IS_DELAYED_INST(nextInstr)){
	    BPatch_reportError(BPatchSerious, 118, "point uninstrumentable");
	    return NULL;
	}
    }

    instruction instr;
    proc->readTextSpace(address, sizeof(instruction), &instr.raw);
	
    //then create the instrumentation point object for the address
    instPoint *newpt = new instPoint((pd_Function *)func,
				     (const instructUnion &)instr,
				     (const image*)(proc->getImage()),
				     (Address &)address,
				     false, // bool delayOk - ignored,
				     otherPoint);

    return proc->findOrCreateBPPoint(bpfunc, newpt, BPatch_instruction);

}

/*
 * BPatch_point::getDisplacedInstructions
 *
 * Returns the instructions to be relocated when instrumentation is inserted
 * at this point.  Returns the number of bytes taken up by these instructions.
 *
 * maxSize      The maximum number of bytes of instructions to return.
 * insns        A pointer to a buffer in which to return the instructions.
 */
int BPatch_point::getDisplacedInstructions(int maxSize, void* insns)
{
    int count = 0;
    instruction copyOut[10];	// I think 7 is the max - jkh 8/3/00

    //
    // This function is based on what is contained in the instPoint 
    //    constructor in the file inst-sparc-solaris.C
    //
    if (!point->hasNoStackFrame()) {
	if (point->ipType == functionEntry) {
	    copyOut[count++].raw = point->saveInsn.raw;
	    copyOut[count++].raw = point->originalInstruction.raw;
	    copyOut[count++].raw = point->delaySlotInsn.raw;
	    if (point->isDelayed) {
		copyOut[count++].raw = point->isDelayedInsn.raw;
		if (point->callAggregate) {
		    copyOut[count++].raw = point->aggregateInsn.raw;
		}
	    }
	} else if (point->ipType == callSite) {
	    copyOut[count++].raw = point->originalInstruction.raw;
	    copyOut[count++].raw = point->delaySlotInsn.raw;
	    if (point->callAggregate) {
		copyOut[count++].raw = point->aggregateInsn.raw;
	    }
	} else {
	    copyOut[count++].raw = point->originalInstruction.raw;
	    copyOut[count++].raw = point->delaySlotInsn.raw;
	}
    } else {
	if (point->ipType == functionEntry) {
	    copyOut[count++].raw = point->originalInstruction.raw;
	    copyOut[count++].raw = point->otherInstruction.raw;
	    copyOut[count++].raw = point->delaySlotInsn.raw;
	    if (point->isDelayed) {
		copyOut[count++].raw = point->isDelayedInsn.raw;
	    }
	} else if (point->ipType == functionExit) {
	    copyOut[count++].raw = point->originalInstruction.raw;
	    copyOut[count++].raw = point->otherInstruction.raw;
	    copyOut[count++].raw = point->delaySlotInsn.raw;
	    if (point->inDelaySlot) {
		copyOut[count++].raw = point->inDelaySlotInsn.raw;
		if (point->firstIsConditional) {
		    copyOut[count++].raw = point->extraInsn.raw;
		}
	    }
	} else if(point->ipType == otherPoint) {
	   copyOut[count++].raw = point->originalInstruction.raw;
	   copyOut[count++].raw = point->otherInstruction.raw;
	} else {
	   assert(point->ipType == callSite);
	   copyOut[count++].raw = point->originalInstruction.raw;
	   copyOut[count++].raw = point->delaySlotInsn.raw;
	   if (point->callAggregate) {
	       copyOut[count++].raw = point->aggregateInsn.raw;
	   }
	}
    }

    if (count * sizeof(instruction) > (unsigned) maxSize) {
	return -1;
    } else {
	memcpy(insns, copyOut, count * sizeof(instruction));
	return count * sizeof(instruction);
    }
}

#endif
